@@ -5,6 +5,7 @@ import collections.abc
 from copy import copy
 from pathlib import Path
 from numbers import Number
+from statistics import mean
 from operator import truediv
 from itertools import chain, repeat, accumulate
 from typing import Any, Dict, List, Tuple, Optional, Sequence, Hashable, Iterator, Union, Type, Set, Callable
@@ -13,7 +14,7 @@ from coba.backports import Literal
 from coba.contexts import CobaContext
 from coba.exceptions import CobaException
 from coba.utilities import PackageChecker
-from coba.pipes import JsonEncode, JsonDecode, DiskIO, ListIO, IO
+from coba.pipes import Pipes, Sink, Source, JsonEncode, JsonDecode, DiskSource, DiskSink, ListSource, ListSink, Foreach
 
 class Table:
     """A container class for storing tabular data."""
@@ -25,7 +26,7 @@ class Table:
             name: The name of the table. Used for display purposes.
             primary_cols: Table columns used to make each row's tuple "key".
             rows: The actual rows that should be stored in the table. Each row is required to contain the given primary_cols.
-            preferred_cols: A list of columns that we prefer be displayed immediately after primary columns. All remaining 
+            preferred_cols: A list of columns that we prefer be displayed immediately after primary columns. All remaining
                 columns (i.e., neither primary nor preferred) will be ordered alphabetically.
         """
         self._name    = name
@@ -43,7 +44,7 @@ class Table:
         col_priority = list(chain(primary_cols + ['index'] + preferred_cols + sorted(all_columns)))
 
         self._columns = sorted(all_columns, key=col_priority.index)
-        self._rows_keys: List[Hashable               ] = []               
+        self._rows_keys: List[Hashable               ] = []
         self._rows_flat: Dict[Hashable, Dict[str,Any]] = {}
         self._rows_pack: Dict[Hashable, Dict[str,Any]] = {}
 
@@ -124,9 +125,7 @@ class Table:
             for col,col_filter in kwargs.items():
 
                 if isinstance(col_filter,collections.abc.Container) and not isinstance(col_filter,str):
-
                     col_filter_results.append(row[col] in col_filter or any([satisifies_filter(cf,row[col]) for cf in col_filter]))
-
                 else:
                     col_filter_results.append(satisifies_filter(col_filter,row.get(col,self._default(col)) ))
 
@@ -174,10 +173,10 @@ class Table:
 
         return pd.DataFrame(col_numpy, columns=self.columns)
 
-    def to_tuples(self) -> Sequence[Tuple[Any,...]]:
+    def to_dicts(self) -> Sequence[Dict[str,Any]]:
         """Turn the Table into a sequence of tuples."""
 
-        tooples = []
+        dicts = []
 
         for key in self.keys:
 
@@ -185,11 +184,12 @@ class Table:
             pack = self._rows_pack[key]
 
             if not pack:
-                tooples.append(tuple(flat.get(col,self._default(col)) for col in self.columns))
+                dicts.append( { col:flat.get(col,self._default(col)) for col in self.columns } )
             else:
-                tooples.extend(list(zip(*[pack.get(col,repeat(flat.get(col,self._default(col)))) for col in self.columns])))
+                tuples = list(zip(*[pack.get(col,repeat(flat.get(col,self._default(col)))) for col in self.columns]))
+                dicts.extend([ dict(zip(self.columns,t)) for t in tuples ])
 
-        return tooples
+        return dicts
 
     def _default(self, column:str) -> Any:
         return [1] if column == "index" else None
@@ -205,7 +205,7 @@ class Table:
                 types.extend([to_type(v) for v in value])
             else:
                 types.append(to_type(value))
-        
+
         return self._resolve_types(types)
 
     def _resolve_types(self, types: Sequence[Optional[Type[Any]]]) -> Type[Union[int,float,bool,object]]:
@@ -213,7 +213,7 @@ class Table:
 
         if len(types) == 1 and types[0] in [dict,str]:
             return object
-        
+
         if len(types) == 1 and types[0] in [int,float,bool]:
             return types[0]
 
@@ -245,12 +245,16 @@ class Table:
 
 class InteractionsTable(Table):
 
-    def to_progressive_lists(self, span: int = None, each: bool = False, ord_col: str = "index", val_col: str = "reward"):
+    def to_progressive_lists(self,
+        span: int = None,
+        each: bool = False,
+        order: str = "index",
+        value: Literal['reward','reward_pct','rank','rank_pct','regret','regret_pct'] = "reward"):
         """Return expanding or exponential averages for col grouped by learner and possibly environment.
 
         Args:
-            span: If span is None return an expanding average (i.e., progressive validation). If span is not
-                none calculate an expontial moving average identical to Pandas df.ewm(span=span).mean().
+            span: If span is None return an expanding average (i.e., progressive validation). If span is not none
+                calculate a simple moving average with window size of span (window will be smaller than span initially).
             each: If true the group by learner and environment (as in each environment). If each is false
                 then only group by learner.
             ord_col: The column which indicates the order in which averaging is calculated on the val_col.
@@ -261,10 +265,17 @@ class InteractionsTable(Table):
         """
         lrn_sim_rows = []
 
+        value_functions = {
+            'reward_pct': lambda interactions: [r/m for r,m in zip(interactions['reward'],interactions['max_reward'])],
+            'rank_pct'  : lambda interactions: [(r-1)/(m-1) for r,m in zip(interactions['rank'],interactions['max_rank'])],
+            'regret'    : lambda interactions: [m-r for r,m in zip(interactions['reward'],interactions['max_reward'])],
+            'regret_pct': lambda interactions: [(u-r)/(u-l) for r,l,u in zip(interactions['reward'],interactions['min_reward'],interactions['max_reward'])],
+        }
+
         for interactions in self:
 
-            values = interactions[val_col]
-            orders = interactions[ord_col]
+            values = value_functions.get(value, lambda interactions: interactions[value])(interactions)
+            orders = interactions[order]
             values = [ v[1] for v in sorted(zip(orders,values)) ]
 
             if span is None or span >= len(values):
@@ -276,9 +287,23 @@ class InteractionsTable(Table):
                 cumdivisor = [1]*len(cumwindow)
 
             else:
-                alpha = 2/(1+span)
-                cumwindow  = list(accumulate(values          , lambda a,c: c + (1-alpha)*a))
-                cumdivisor = list(accumulate([1.]*len(values), lambda a,c: c + (1-alpha)*a)) #type: ignore
+                moving_sum = 0
+                cumwindow  = []
+                cumdivisor = []
+                for index in range(len(values)):
+                    sub_i = index-span
+                    add_i = index
+
+                    moving_sum += values[add_i]
+                    moving_sum -= values[sub_i] if sub_i >= 0 else 0
+
+                    cumwindow.append(moving_sum)
+                    cumdivisor.append(min(span,add_i+1))
+
+                #highly-performant way to calucate exponential moving average identical to Pandas df.ewm(span=span).mean()
+                #alpha = 2/(1+span)
+                #cumwindow  = list(accumulate(values          , lambda a,c: c + (1-alpha)*a))
+                #cumdivisor = list(accumulate([1.]*len(values), lambda a,c: c + (1-alpha)*a)) #type: ignore
 
             lrn_sim_rows.append([interactions["learner_id"], interactions["environment_id"], *list(map(truediv, cumwindow, cumdivisor))])
 
@@ -309,8 +334,8 @@ class InteractionsTable(Table):
         """Return expanding or exponential averages for yaxis grouped by learner and possibly environment.
 
         Args:
-            span: If span is None return an expanding average (i.e., progressive validation). If span is not
-                none calculate an expontial moving average identical to Pandas df.ewm(span=span).mean().
+            span: If span is None return an expanding average (i.e., progressive validation). If span is not none
+                calculate a simple moving average with window size of span (window will be smaller than span initially).
             each: If true the group by learner and environment (as in each environment). If each is false
                 then only group by learner.
             ord_col: The column which indicates the order in which averaging is calculated on the val_col.
@@ -334,19 +359,21 @@ class InteractionsTable(Table):
             n_index = len(data[0][1:])
             return pd.DataFrame(data, columns=["learner_id", *range(1,n_index+1)])
 
-class TransactionIO_V3(IO['Result', Any]):
+class TransactionIO_V3(Source['Result'], Sink[Any]):
 
-    def __init__(self, transaction_log: Optional[str] = None, minify:bool=True) -> None:
+    def __init__(self, log_file: Optional[str] = None, minify:bool = True) -> None:
 
-        self._io = DiskIO(transaction_log) if transaction_log else ListIO()
-        self._minify = minify
+        self._log_file = log_file
+        self._minify   = minify
+        self._source   = DiskSource(log_file) if log_file else ListSource()
+        self._sink     = DiskSink(log_file) if log_file else ListSink(self._source.items)
 
     def write(self, item: Any) -> None:
-        if isinstance(self._io, ListIO):
-            self._io.write(self._encode(item))
-        else: 
-            if not Path(self._io._filename).exists():self._io.write('["version",3]')
-            self._io.write(JsonEncode(self._minify).filter(self._encode(item)))
+        if isinstance(self._sink, ListSink):
+            self._sink.write(self._encode(item))
+        else:
+            if not Path(self._sink._filename).exists():self._sink.write('["version",3]')
+            self._sink.write(JsonEncode(self._minify).filter(self._encode(item)))
 
     def read(self) -> 'Result':
         n_lrns   = None
@@ -355,21 +382,26 @@ class TransactionIO_V3(IO['Result', Any]):
         sim_rows = {}
         int_rows = {}
 
-        for trx in self._io.read() if isinstance(self._io, ListIO) else map(JsonDecode().filter,self._io.read()):
+        if isinstance(self._source, ListSource):
+            decoded_source = self._source
+        else:
+            decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))
+
+        for trx in decoded_source.read():
 
             if not trx: continue
 
-            if trx[0] == "benchmark": 
+            if trx[0] == "benchmark":
                 n_lrns = trx[1]["n_learners"]
                 n_sims = trx[1]["n_simulations"]
-            
-            if trx[0] == "S": 
+
+            if trx[0] == "S":
                 sim_rows[trx[1]] = trx[2]
-            
-            if trx[0] == "L": 
+
+            if trx[0] == "L":
                 lrn_rows[trx[1]] = trx[2]
-            
-            if trx[0] == "I": 
+
+            if trx[0] == "I":
                 int_rows[tuple(trx[1])] = trx[2]
 
         return Result(n_lrns, n_sims, sim_rows, lrn_rows, int_rows)
@@ -397,18 +429,21 @@ class TransactionIO_V3(IO['Result', Any]):
 
         return None
 
-class TransactionIO_V4(IO['Result', Any]):
+class TransactionIO_V4(Source['Result'], Sink[Any]):
 
-    def __init__(self, transaction_log: Optional[str] = None, minify:bool=True) -> None:
-        self._io     = DiskIO(transaction_log) if transaction_log else ListIO()
-        self._minify = minify
+    def __init__(self, log_file: Optional[str] = None, minify:bool=True) -> None:
+
+        self._log_file = log_file
+        self._minify   = minify
+        self._source   = DiskSource(log_file) if log_file else ListSource()
+        self._sink     = DiskSink(log_file)   if log_file else ListSink(self._source.items)
 
     def write(self, item: Any) -> None:
-        if isinstance(self._io, ListIO):
-            self._io.write(self._encode(item))
-        else: 
-            if not Path(self._io._filename).exists():self._io.write('["version",4]')
-            self._io.write(JsonEncode(self._minify).filter(self._encode(item)))
+        if isinstance(self._sink, ListSink):
+            self._sink.write(self._encode(item))
+        else:
+            if not Path(self._sink._filename).exists():self._sink.write('["version",4]')
+            self._sink.write(JsonEncode(self._minify).filter(self._encode(item)))
 
     def read(self) -> 'Result':
 
@@ -418,21 +453,26 @@ class TransactionIO_V4(IO['Result', Any]):
         env_rows = {}
         int_rows = {}
 
-        for trx in self._io.read() if isinstance(self._io, ListIO) else map(JsonDecode().filter,self._io.read()):
+        if isinstance(self._source, ListSource):
+            decoded_source = self._source
+        else:
+            decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))
+
+        for trx in decoded_source.read():
 
             if not trx: continue
 
-            if trx[0] == "experiment": 
+            if trx[0] == "experiment":
                 n_lrns = trx[1]["n_learners"]
                 n_sims = trx[1]["n_environments"]
-            
-            if trx[0] == "E": 
+
+            if trx[0] == "E":
                 env_rows[trx[1]] = trx[2]
-            
-            if trx[0] == "L": 
+
+            if trx[0] == "L":
                 lrn_rows[trx[1]] = trx[2]
-            
-            if trx[0] == "I": 
+
+            if trx[0] == "I":
                 int_rows[tuple(trx[1])] = trx[2]
 
         return Result(n_lrns, n_sims, env_rows, lrn_rows, int_rows)
@@ -460,14 +500,14 @@ class TransactionIO_V4(IO['Result', Any]):
 
         return None
 
-class TransactionIO(IO['Result', Any]):
+class TransactionIO(Source['Result'], Sink[Any]):
 
     def __init__(self, transaction_log: Optional[str] = None) -> None:
 
         if not transaction_log or not Path(transaction_log).exists():
             version = None
         else:
-            version = JsonDecode().filter(next(DiskIO(transaction_log).read()))[1]
+            version = JsonDecode().filter(next(DiskSource(transaction_log).read()))[1]
 
         if version == 3:
             self._transactionIO = TransactionIO_V3(transaction_log)
@@ -494,7 +534,7 @@ class Result:
     def from_file(filename: str) -> 'Result':
         """Create a Result from a transaction file."""
 
-        if not Path(filename).exists(): 
+        if not Path(filename).exists():
             raise CobaException("We were unable to find the given Result file.")
 
         return TransactionIO(filename).read()
@@ -526,33 +566,33 @@ class Result:
 
     @property
     def learners(self) -> Table:
-        """The collection of learners used in the Experiment. 
-        
+        """The collection of learners used in the Experiment.
+
         The primary key of this table is learner_id.
         """
         return self._learners
 
     @property
     def environments(self) -> Table:
-        """The collection of environments used in the Experiment. 
-        
+        """The collection of environments used in the Experiment.
+
         The primary key of this table is environment_id.
         """
         return self._environments
 
     @property
     def interactions(self) -> InteractionsTable:
-        """The collection of interactions that learners chose actions for in the Experiment. 
-            
-        The primary key of this Table is (index, environment_id, learner_id). It should be noted that the InteractionTable 
-        has an additional property, to_progressive_pandas() which calculates the expanding moving average or exponential 
+        """The collection of interactions that learners chose actions for in the Experiment.
+
+        The primary key of this Table is (index, environment_id, learner_id). It should be noted that the InteractionTable
+        has an additional property, to_progressive_pandas() which calculates the expanding moving average or exponential
         moving average for interactions ordered by index and grouped by environment_id and learner_id.
         """
         return self._interactions
 
     def copy(self) -> 'Result':
         """Create a copy of Result."""
-        
+
         result = Result()
 
         result._environments = copy(self._environments)
@@ -561,20 +601,32 @@ class Result:
 
         return result
 
-    def filter_fin(self) -> 'Result':
-        """Filter the result to only contain data about environments with all learners.
+    def filter_fin(self, n_interactions: int = None) -> 'Result':
+        """Filter the result to only contain data about environments with all learners and interactions.
 
         Args:
-            pred: A predicate that returns true for learner dictionaries that should be kept.
-            **kwargs: key-value pairs to filter on. For more information see Table.filter.
+            n_interactions: The number of interactions at which an environment is considered complete.
         """
 
-        def is_complete_sim(sim_id):
-            return all((sim_id, lrn_id) in self.interactions for lrn_id in self.learners.keys)
+        def has_all(env_id):
+            return all((env_id, lrn_id) in self.interactions for lrn_id in self.learners.keys)
+
+        def has_min(env_id):
+            if not n_interactions:
+                return True
+            else:
+                return all(len(self.interactions[(env_id, lrn_id)].get('index',[])) >= n_interactions for lrn_id in self.learners.keys)
+
+        complete_ids = [env_id for env_id in self.environments.keys if has_all(env_id) and has_min(env_id)]
 
         new_result               = copy(self)
-        new_result._environments = self.environments.filter(environment_id=is_complete_sim)
-        new_result._interactions = self.interactions.filter(environment_id=is_complete_sim)
+        new_result._environments = self.environments.filter(environment_id=complete_ids)
+        new_result._interactions = self.interactions.filter(environment_id=complete_ids)
+
+        if n_interactions:
+            new_inter = copy(new_result.interactions)
+            new_inter._rows_pack = { rk: { pk: pv[0:n_interactions] for pk, pv in rv.items() } for rk,rv in new_inter._rows_pack.items() }
+            new_result._interactions = new_inter
 
         if len(new_result.environments) == 0:
             CobaContext.logger.log(f"There was no environment which was finished for every learner.")
@@ -615,28 +667,32 @@ class Result:
 
         return new_result
 
-    def plot_learners(self, 
+    def plot_learners(self,
+        y    : Literal['reward','reward_pct','rank','rank_pct','regret','regret_pct'] = "reward",
         xlim : Optional[Tuple[Number,Number]] = None,
         ylim : Optional[Tuple[Number,Number]] = None,
         span : int = None,
         err  : Optional[Literal['se','sd']] = None,
         each : bool = False,
         filename: str = None,
+        sort : Literal["name","id","y"] = "y",
         ax = None) -> None:
-        """Plot the performance of multiple learners on multiple environments. It gives a sense of the expected 
-            performance for different learners across independent environments. This plot is valuable in gaining 
-            insight into how various learners perform in comparison to one another. 
+        """Plot the performance of multiple learners on multiple environments. It gives a sense of the expected
+            performance for different learners across independent environments. This plot is valuable in gaining
+            insight into how various learners perform in comparison to one another.
 
         Args:
+            y: The value to plot on the y-axis: `reward` plots the reward obtained on each interaction, `pct_reward`
+                plots the percentage of the optimal possible reward obtained on each interaction.
             xlim: Define the x-axis limits to plot. If `None` the x-axis limits will be inferred.
             ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
-            span: In general this indicates how many previous evaluations to average together. In practice this works
-                identically to ewm span value in the Pandas API. Additionally, if span equals None then all previous 
-                rewards are averaged together vs span = 1 WHERE the instantaneous reward is plotted for each interaction.
+            span: If span is None return an expanding average (i.e., progressive validation). If span is not none
+                calculate a simple moving average with window size of span (window will be smaller than span initially).
             err: This determines what kind of error bars to plot (if any). Valid types are `None`, 'se', and 'sd'. If `None`
                 then no bars are plotted, if 'se' the standard error is shown, and if 'sd' the standard deviation is shown.
             each: This determines whether each evaluated environment used to estimate mean performance is also plotted.
             filename: Provide a filename to write plot image to disk.
+            sort: Indicate how learner names should be sorted in the legend.
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
 
@@ -644,8 +700,9 @@ class Result:
         import matplotlib.pyplot as plt #type: ignore
 
         show = ax is None
+        n_environments = []
 
-        for label, X, Y, yerr, Z in self._plot_learners_data(xlim,span,err):
+        for label, X, Y, yerr, Z in self._plot_learners_data(y,xlim,span,err,sort):
 
             ax = ax or plt.figure(figsize=(10,6)).add_subplot(111) #type: ignore
 
@@ -653,6 +710,8 @@ class Result:
 
             ax.errorbar(X, Y, yerr=yerr, elinewidth=0.5, errorevery=(0,max(int(len(X)*0.05),1)), label=label, color=color)
 
+            n_environments.append(len(Z[0]))
+
             if each:
                 for Y in list(zip(*Z)):
                     ax.plot(X,Y, color=color, alpha=0.15)
@@ -673,8 +732,10 @@ class Result:
                 y_pad = padding*(ylim[1]-ylim[0])
                 ax.set_ylim(ylim[0]-y_pad, ylim[1]+y_pad)
 
-            ax.set_title(("Instantaneous" if span == 1 else "Progressive" if span is None else f"Span {span}") + " Reward", loc='left',pad=15)
-            ax.set_ylabel("Reward")
+            y_label = y.capitalize().replace("_pct"," Percent")
+
+            ax.set_title(("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {y_label} ({mean(n_environments)} Environments)", loc='left',pad=15)
+            ax.set_ylabel(y_label)
             ax.set_xlabel("Interactions")
 
             if ax.get_legend() is None:
@@ -694,19 +755,42 @@ class Result:
                 plt.savefig(filename, dpi=300)
                 plt.close()
 
-    def _plot_learners_data(self, xlim: Tuple[Number,Number] = None, span: int = None, err: Literal['se','sd'] = None):
+    def _plot_learners_data(self,
+        y   : Literal['reward','reward_pct','rank','rank_pct','regret','regret_pct'] = "reward",
+        xlim: Tuple[Number,Number] = None,
+        span: int = None,
+        err : Literal['se','sd'] = None,
+        sort: Literal['name',"id","y"] = "y"):
 
         if xlim and xlim[0] >= xlim[1]:
-            CobaContext.logger.log("The given x-limit end is less than the x-limit start. Plotting is impossible.")
+            CobaContext.logger.log("The xlim end is less than the xlim start. Plotting is impossible.")
 
+        first_length = None
+        mixed_length = False
         progressives: Dict[int,List[Sequence[float]]] = collections.defaultdict(list)
 
-        for progressive in self.interactions.to_progressive_lists(span=span, each=True, ord_col="index", val_col="reward"):
-            progressives[progressive[0]].append(progressive[2:])
+        for progressive in self.interactions.to_progressive_lists(span=span, each=True, order="index", value=y):
+            if not xlim or not xlim[1] or len(progressive[2:]) >= xlim[1]:
+                progressives[progressive[0]].append(progressive[2:])
+                first_length = first_length or len(progressive[2:])
+            if (not xlim or not xlim[1]) and first_length and len(progressive[2:]) != first_length:
+               mixed_length = True
+
+        if mixed_length:
+            CobaContext.logger.log("The result contains environments of different lengths. The plot only includes data which is present in all environments. To only plot environments with a minimum number of interactions call <result>.filter_fin(n_interactions).")
 
         if progressives and (not xlim or xlim[0] < xlim[1]):
 
-            for learner_id in sorted(self.learners.keys, key=lambda id: self.learners[id]["full_name"]):
+            if sort == "name":
+                sort_func = lambda id: self.learners[id]["full_name"]
+
+            if sort == "y":
+                sort_func = lambda id: -sum(list(zip(*progressives[id]))[-1])
+
+            if sort == "id":
+                sort_func = lambda id: id
+
+            for learner_id in sorted(self.learners.keys, key=sort_func):
 
                 label = self._learners[learner_id]["full_name"]
                 Z     = list(zip(*progressives[learner_id]))
@@ -717,8 +801,7 @@ class Result:
                 Y     = [ sum(z)/len(z) for z in Z ]
                 X     = list(range(1,len(Y)+1))
 
-                start = xlim[0] if xlim else int(.05*len(X))
-                end   = xlim[1] if xlim else len(X)
+                start,end = xlim if xlim else (0,len(X))
 
                 X = X[start:end]
                 Y = Y[start:end]
