@@ -10,7 +10,7 @@ from coba.learners.primitives import Learner, Probs, Info
 class EpsilonBanditLearner(Learner):
     """A bandit learner using epsilon-greedy for exploration."""
 
-    def __init__(self, epsilon: float) -> None:
+    def __init__(self, epsilon: float=.05) -> None:
         """Instantiate an EpsilonBanditLearner.
 
         Args:
@@ -34,7 +34,7 @@ class EpsilonBanditLearner(Learner):
         prob_selected_randomly = [1/len(actions) * self._epsilon] * len(actions)
         prob_selected_greedily = [ int(i in max_indexes)/len(max_indexes) * (1-self._epsilon) for i in range(len(actions))]
 
-        return [ round(p1+p2,4) for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]
+        return [ p1+p2 for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]
 
     def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
 
@@ -47,18 +47,18 @@ class EpsilonBanditLearner(Learner):
 
 class UcbBanditLearner(Learner):
     """A bandit learner using upper confidence bound estimates for exploration.
-    
+
     This algorithm is an implementation of Auer et al. (2002) UCB1-Tuned algorithm
     and requires that all rewards are in [0,1].
 
     References:
-        Auer, Peter, Nicolo Cesa-Bianchi, and Paul Fischer. "Finite-time analysis of 
+        Auer, Peter, Nicolo Cesa-Bianchi, and Paul Fischer. "Finite-time analysis of
         the multiarmed bandit problem." Machine learning 47.2-3 (2002): 235-256.
     """
-    
+
     def __init__(self):
         """Instantiate a UcbBanditLearner."""
-        #these variable names were selected for easier comparison with the original paper 
+        #these variable names were selected for easier comparison with the original paper
         self._t     : int = 0
         self._m     : Dict[Action, float         ] = {}
         self._s     : Dict[Action, int           ] = {}
@@ -138,7 +138,7 @@ class FixedLearner(Learner):
         Args:
             fixed_pmf: A PMF whose values are the probability of taking each action.
         """
-        
+
         assert round(sum(fixed_pmf),3) == 1, "The given pmf must sum to one to be a valid pmf."
         assert all([p >= 0 for p in fixed_pmf]), "All given probabilities of the pmf must be greater than or equal to 0."
 
@@ -150,7 +150,7 @@ class FixedLearner(Learner):
         return {"family":"fixed"}
 
     def predict(self, context: Context, actions: Sequence[Action]) -> Probs:
- 
+
         return self._fixed_pmf
 
     def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
